

/* The assignment follows a similar approach as practised in the lab sessions.
 * The program executes some tests [tests 1-5] which should fail on
 * providing the missing contracts. The assignment requires addition of 
 * contracts to the class "Library". The comments below in the class definition
 * provides information about the methods and data structure.
 * 
 * Please feel free to use the protected methods as helper functions to 
 * write contracts. We may not add contracts to these functions. We can assume
 * that none of the parameters are null or empty for the assignment.
 * 
 * The contracts are supposed to meet some basic contraints as mentioned below:
 * 
 * 1. We must not add duplicate members or copies of book with duplicate IDs
 * 2. Only members can borrow available books.
 * 3. A member cannot borrow multiple copies of the same book.
 * 4. Number of available copies of any book must not be less than 1 (i.e the 
 * 	library must have atleast one copy of each books available at all times say 
 * 	for the reading room) and the total number of copies of a books should not 
 * 	exceed the number of members. This property must satisfy at all times.
 * 5. The methods must ensure their objectives are met.
 * 
 * 
 * To run the tests, please use the "testrun" script from command line from its
 * directory as ".\testrun.bat <number>". If <number> is 1 to 5, it will run the 
 * specific test and the number 0 will run all tests. For example if you have
 * extracted the contents of the folder to directory "ITI0130_Lab7_assignment_cofoja" and want
 * to run test 4, then use the commands: 
 * 
 * > cd <path-to-ITI0130_Lab7_assignment_cofoja>
 * > .\testrun.bat 4
 * 
 * In Linux, use "sh ./testrun.sh 4" instead.
 * 
 * Initially all test should pass, with no exceptions. We need to add contracts
 * to fail the tests.
 * 
 */


import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;

import com.google.java.contract.Ensures;
import com.google.java.contract.Invariant;
import com.google.java.contract.Requires;
import com.google.java.contract.ThrowEnsures;


/*
 * The Library class implements a highly simplified and limited model of 
 * a library. LibraryTest class executes some tests on Library.
 * 
 */
class Library {


	/*
	 * The constructor initialize the data structures of Library class
	 * The data consists of two collections. 
	 * 
	 * collection - Map<Pair<String,String>, Map<String, Boolean>>
	 * 
	 * "collection" is the structure which stores the book data. It is a nested 
	 * map that stores books as <name, author> Pairs and each book is mapped to 
	 * another map that stores aiailability of copies (IDs) of the book as 
	 * <id (String), availability (Boolean)>. Each book has multiple copies with
	 * unique ids.
	 * 
	 * members - Map<String, ArrayList<String>> 
	 * 
	 * The collection "members" is a map that stores the names of the member
	 * of the library and maps with each name a list of books IDs they borrowed.
	 * 
	 */
	public Library() {
		
		collection = new HashMap<Pair<String,String>,Map<String, Boolean>>();
		members = new HashMap<String, ArrayList<String>>();

		/*
		 * The initialize function adds 9 books ( 3 books and 3 copies each ) and
		 * 3 members to the library.
		 */
		initialize();
	}

	/*
	 * The function accepts the name of the member as argument and add it to 
	 * the list. 
	 */
	public void addMember(String name) {
		members.put(name, new ArrayList<String>());
	}

	/*
	 * Adds a books to the library. The function accepts name , author and ID 
	 * of the book. If the book(name, author) already exists, it adds the new ID
	 * to list of copies of the book and marks it as available.
	 */
	public void addBook(String name, String author, String id) {

		Pair<String,String> b = getBook(name, author);

		if(!collection.containsKey(b)) {
			Map<String, Boolean> idmap = new HashMap<String, Boolean>();
			collection.put(b, idmap);
		}

		// Add the copy to the list and mark it available.
		collection.get(b).put(id, true);
	}

	/*
	 * The funtion accepts member name, book name and author as arguments
	 * and marks the book as borrowd. The book is borrowed by adding the 
	 * ID to the list of borrowed book by the member and marking the copy as
	 * unavailable in the library book data.
	 */
	public void borrowbook(String member, String name, String author) {

		Pair<String,String> b = getBook(name, author);
		Iterator<String> it = collection.get(b).keySet().iterator();
		String id;

		do { 
			id = it.next(); 
		} while (!collection.get(b).get(id));

		// Add the copy to the members borrowed list
		members.get(member).add(id);
		// Mark the copy as unavailable.
		collection.get(b).put(id, false);

	}

	/*
	 * Helper function to check the borrowed condition. i) Member has a copy of 
	 * the book. ii) The same copy is marked unavailable in the collection.
	 */
	protected boolean isBorrowed(String member, String name, String author) {

		boolean br = false;
		Pair<String,String> b = getBook(name, author);
		Iterator<String> it = collection.get(b).keySet().iterator();

		while(it.hasNext()) {

			String id = it.next();
			if(members.get(member).contains(id) && !collection.get(b).get(id)) {
				br = true;
				break;
			}	
		}

		return br;
	}

	/*
	 * The function check the validity of two conditions which satisfy constraint 4
	 * of the assignment. It check for i) at least one copy of each book is available
	 * and ii) the number of copies of a book does not exceed the number of members.
	 * The function returns true of the conditions are satisfied.
	 */
	protected boolean isManaged() {
		boolean br = true;
		Iterator<Map<String, Boolean>> it = collection.values().iterator();

		while(it.hasNext()) {

			Map<String, Boolean> idmap = it.next();

			if(!idmap.containsValue(true) || idmap.size() > members.size()) {
				br = false;
				break;
			}
		}

		return br;
	}

	/*
	 * Checks the availability of a book
	 */
	protected boolean isBookAvailable(String name, String author) {

		Pair<String,String> b = getBook(name, author);
		return (collection.containsKey(b) && collection.get(b).containsValue(true));
	}

	/*
	 * Checks if member already exists.
	 */
	protected boolean memberExists(String member) {
		return members.containsKey(member);
	}

	/*
	 * Checks the availability of a specific copy of a book
	 */
	protected boolean isBookAvailable(String name, String author, String id) {

		Pair<String,String> b = getBook(name, author);
		return (collection.containsKey(b) && collection.get(b).containsKey(id));
	}

	/*
	 * Checks if the ID already exists in the collection.
	 */
	protected boolean isDuplicateID(String id) {

		boolean br = false;
		Iterator<Map<String, Boolean>> it = collection.values().iterator();

		while(it.hasNext()) {
			if(it.next().keySet().contains(id)) {
				br = true;
				break;
			}
		}

		return br;
	}

	protected Pair<String,String> getBook(String name, String author) {
		return new ImmutablePair<String,String>(name,author);
	}

	protected void initialize() {

		// Add books to Library
		for(Integer i = 0 ; i < 3 ; i++) {
			String name = "book" + i.toString();
			String author = "author" + i.toString();
			for( Integer j = 0 ; j < 3 ; j++) {
				String id = "id" + i.toString() + j.toString();
				addBook(name, author, id);
			}
		}

		//Add members to Library
		for(Integer i = 0 ; i < 3 ; i++)
			addMember("member" + i.toString());
	}

	private Map<String, ArrayList<String>> members;
	private Map<Pair<String,String>, Map<String, Boolean>> collection;
}


/*
 * The LibraryTest class executes some tests on the Library class
 * The tests are performed by the test<number>() functions called 
 * from main.
 * 
 */

public class LibraryTest {

	public LibraryTest() {
		lib = new Library();
	}


	public void test1() {
		lib.addMember("member1");
	}

	public void test2() {
		lib.addBook("book1", "author1", "id11");
	}

	public void test3() {
		lib.borrowbook("member0", "book0", "author0");
		lib.borrowbook("member0", "book0", "author0");
	}

	public void test4() {
		lib.borrowbook("member0", "book0", "author0");
		lib.borrowbook("member1", "book0", "author0");
		lib.borrowbook("member2", "book0", "author0");
	}

	public void test5() {
		lib.addBook("book1", "author1", "id13");
	}

	public static void main(String[] args) {

		LibraryTest lt = new LibraryTest();

		if(args.length != 1)
			System.out.println("Please provide test number[1-5]");
		else {

			try {

				Integer tn = Integer.parseInt(args[0]);
				
				if(tn >= 1 && tn <= 5)
				{
					System.out.print("Executing test " + tn + " ==> ") ;
					
					switch(tn) {
					case 1: lt.test1();
					break;
					case 2: lt.test2();
					break;
					case 3: lt.test3();
					break;
					case 4: lt.test4();
					break;
					case 5: lt.test5();
					break;
					default:
						break;
					}
					
					System.out.println("OK");
				}
				else
					System.out.println("Please provide test number[1-5]");

			} catch (Exception e) {
				System.out.println("Please provide test number[1-5]");
			}


		}
	}


	private Library lib;


}
